#include "Image.h"
#include "ImageDiskXFile.h"

namespace Torch {

Image::Image()
{
	addBOption("verbose", &verbose, false, "verbose");

	image_id = -1;

   	data = NULL;
	pixmap = NULL;
	
	width = 0;
	height = 0;
	image_size = 0;
	real_image_size = 0;
	
	n_planes = 0;
	coding = NULL;
}

Image::Image(int width_, int height_, int n_planes_) : Sequence(1, width_ * height_ * n_planes_)
{
	addBOption("verbose", &verbose, false, "verbose");

	image_id = -1;

	width = width_;
	height = height_;
	n_planes = n_planes_;
	image_size = width_ * height_;
	real_image_size = image_size;

	data = frames[0];
	pixmap = (unsigned char *) allocator->alloc(image_size * n_planes * sizeof(unsigned char));

	for(int i = 0 ; i < image_size * n_planes ; i++)
	{
		data[i] = 0.0;
		pixmap[i] = 0;
	}

	coding = NULL;
}

void Image::updateDataFromPixmap()
{
	for(int i = 0 ; i < image_size * n_planes ; i++) data[i] = (real) pixmap[i];
}

void Image::updatePixmapFromData()
{
	for(int i = 0 ; i < image_size * n_planes ; i++)
	{
	   	real pixel = data[i];

		if(pixel < 0.0) pixel = 0;
		if(pixel > 255.0) pixel = 255.0;

		pixmap[i] = FixI(pixel);
	}
}

bool Image::saveImageXFile(ImageDiskXFile *file)
{
   	if(n_planes != file->n_planes) error("Image::saveImageXFile() incompatible n_planes");
	   
	file->writeHeader(width, height);
	file->writePixmap(pixmap);

	return true;
}

bool Image::loadImageXFile(ImageDiskXFile *file)
{
   	//if(n_planes != file->n_planes) error("Image::loadImageXFile() incompatible n_planes");
	   
   	// reads header
	if(!file->readHeader(&width, &height)) return false;

	image_size = width * height;
	frame_size = image_size;

	// allocate/reallocate memory if needed
	imageRealloc();

	// reads image data
	file->readPixmap();

	// copy/convert
	copyFrom(width, height, file->pixmap, file->coding);
	
	// update image data from the pixmap
	updateDataFromPixmap();

	return true;
}

void Image::saveXFile(XFile *file)
{
}

void Image::loadXFile(XFile *file)
{
}

void Image::imageRealloc()
{
	if(pixmap == NULL)
	{
	   	//
		pixmap = (unsigned char *) allocator->alloc(n_planes * image_size * sizeof(unsigned char));
		real_image_size = image_size;

		//
		n_real_frames = 1;
		n_frames = 1;
		frame_size = image_size;
		frames = (real **)allocator->alloc(n_frames * sizeof(real *));
		frames[0] = (real *)allocator->alloc(n_planes * frame_size * sizeof(real));
		data = frames[0];
	}
	else
	{
	   	if(image_size > real_image_size)
		{
			pixmap = (unsigned char *) allocator->realloc((unsigned char *) pixmap, n_planes * image_size * sizeof(unsigned char));
			real_image_size = image_size;

			frames[0] = (real *)allocator->realloc((real *) frames[0], n_planes * image_size * sizeof(real));
			data = frames[0];
		}
	}
	
}

void Image::drawline(int x1, int y1, int x2, int y2, Color c)
{
	// THE EXTREMELY FAST LINE ALGORITHM Variation E (Addition Fixed Point PreCalc)
	// See original code in FastLine.cc
	 
   	int x = x1;
	int y = y1;

        bool yLonger=false;
        int shortLen=y2-y;
        int longLen=x2-x;
        if (abs(shortLen)>abs(longLen)) 
	{
                int swap=shortLen;
                shortLen=longLen;
                longLen=swap;                           
                yLonger=true;
        }
        int decInc;
        if (longLen==0) decInc=0;
        else decInc = (shortLen << 16) / longLen;

        if (yLonger) 
	{
                if (longLen>0) 
		{
                        longLen+=y;
                        for (int j=0x8000+(x<<16);y<=longLen;++y) 
			{
                                drawpixel(j >> 16,y, c);     
                                j+=decInc;
                        }
                        return;
                }
                longLen+=y;
                for (int j=0x8000+(x<<16);y>=longLen;--y) 
		{
                        drawpixel(j >> 16,y,c);     
                        j-=decInc;
                }
                return; 
        }

        if (longLen>0) 
	{
                longLen+=x;
                for (int j=0x8000+(y<<16);x<=longLen;++x) 
		{
                        drawpixel(x,j >> 16,c);
                        j+=decInc;
                }
                return;
        }
        longLen+=x;
        for (int j=0x8000+(y<<16);x>=longLen;--x) 
	{
                drawpixel(x,j >> 16,c);
                j-=decInc;
        }
}

void Image::drawline(double x1, double y1, double x2, double y2, Color c)
{
	drawline((int) x1, (int) y1, (int) x2, (int) y2, c);
}

Image::~Image()
{
}

}
