#include "ipDCTmod2.h"

namespace Torch {

ipDCTmod2::ipDCTmod2(int width_, int height_, const char *coding_, int dct_dim, int block_size_, int block_overlap_, int delta_neighbour_, int delta_keep_) : ipCore(coding_)
{
	width = width_;
        height = height_;

	block_size = block_size_;
	block_overlap = block_overlap_;
	
	if(block_size < 1)
		error("ipDCTmod2: block size (%d) incorrect.", block_size);   	
	
	if((block_overlap < 0) || (block_overlap > (block_size-1)))
		error("ipDCTmod2: overlap should be in range [0...%d].", block_size-1);   	
	delta_neighbour = delta_neighbour_;
	delta_keep = delta_keep_;

	ip = new(allocator) ipDCT2D(block_size, "gray", dct_dim);
	ip_output_size = dct_dim;
	output_size = dct_dim + delta_keep;
	
	// set intern tmp variables
	init_();
	
	seq_out = new(allocator) Sequence(n_block, output_size);
}

void ipDCTmod2::init_()
{
	pixeladv = block_size - block_overlap;

	rowblocks = 0;
        colblocks = 0;
        int row_, col_;

	for(row_ = 1 ; row_ <= height-(block_size-1) ; row_ += pixeladv)
		rowblocks++;

	for(col_ = 1 ; col_ <= width-(block_size-1) ; col_ += pixeladv)
		colblocks++;

	int end_row = row_ - pixeladv + (block_size-1);
	int end_col = col_ - pixeladv + (block_size-1);

	int rem_rows = height - end_row;
	int rem_cols = width - end_col;

	row_offset = rem_rows/2;
	col_offset = rem_cols/2;

	n_block = rowblocks * colblocks;
	
	subimage = new(allocator) Sequence(1, n_planes*block_size*block_size);

	n_block_in = n_block;
	rowblocks_in = rowblocks;
	colblocks_in = colblocks;
	
	tmp = (real *)allocator->alloc(output_size*sizeof(real));
	dct = (real *)allocator->alloc(n_block_in*ip_output_size*sizeof(real));
	coldelta = (real *)allocator->alloc(n_block_in*ip_output_size*sizeof(real));
	rowdelta = (real *)allocator->alloc(n_block_in*ip_output_size*sizeof(real));
	
	// Output blocks: -2*delta_neighbour (we're not considering deltas for edges)
	rowblocks = rowblocks_in-2*delta_neighbour;
	colblocks = colblocks_in-2*delta_neighbour;
	n_block = rowblocks*colblocks;

	if(rowblocks < 1) 
		error("ipDCTmod2::init() not possible for given image dimensions."); 
	
	if(colblocks < 1)
		error("ipDCTmod2::init() not possible for given image dimensions."); 
	
	if(rowblocks < 1) 
	   	warning("rowblockst < 1; suggestion to use a larger image.");
	
	if(colblocks < 1)
		warning("colblocks < 1; suggestion to use a larger image.");

	if(verbose)
	{
		print("ipDCTmod2::init()\n");
		print("   row_offset = %d\n", row_offset);
		print("   col_offset = %d\n", col_offset);
		print("   rows = %d\n", height);
		print("   cols = %d\n", width);
		print("   block size = %d\n", block_size);
                print("   block overlap = %d\n", block_overlap);
		print("   delta neighbour = %d\n", delta_neighbour);
		print("   delta keep = %d\n", delta_keep);
		print("   ip (DCT) output_size = %d\n", ip_output_size);
		print("");	
		print("   n_block in = %d\n", n_block_in);
		print("   rowblocks in = %d\n", rowblocks_in);
		print("   colblocks in = %d\n", colblocks_in);
		print("");	
	   	print("   DCTmod2 n_block out = %d\n", n_block);
		print("   DCTmod2 rowblocks out = %d\n", rowblocks);
		print("   DCTmod2 colblocks out = %d\n", colblocks);
	   	print("   DCTmod2 output_size = %d\n", output_size);
	}
		
}

void ipDCTmod2::process(Sequence *seq_in)
{
	for(int rowblock=0; rowblock<rowblocks_in; rowblock++)
	{
	   	for(int colblock=0; colblock<colblocks_in; colblock++) 
		{
			int row = row_offset + rowblock*pixeladv;
			int col = col_offset + colblock*pixeladv;

			// Extract a block of the input image
			switch(coding_id)
			{
			case 0:
				cropGray(seq_in->frames[0], width, height, col, row, block_size, block_size, subimage->frames[0]);
				break;
			case 1:
				cropRgb(seq_in->frames[0], width, height, col, row, block_size, block_size, subimage->frames[0]);
				break;
			case 2:
				message("ipDCTmod2::process() not implemented for YUV format");
				break;
			case 3:
				message("ipDCTmod2::process() not implemented for float format");
				break;
			}

			// process the block 
			ip->process(subimage);

			// Storing blocks as [colblock][rowblock] (i.e. [x][y])
		   	int index_block = colblock + rowblock * colblocks_in;

			// Copy the output of the pattern machine
			real *output_fim_ = ip->seq_out->frames[0];
		   	real *output_block_ = &dct[index_block*ip_output_size];
		
			for(int i=0 ; i < ip_output_size ; i++) 
				output_block_[i] = output_fim_[i];
		}
	}

	// horizontal delta
	for(int rowblock=0; rowblock<rowblocks_in; rowblock++)
	{
	   	for(int colblock=delta_neighbour; colblock<(colblocks_in-delta_neighbour); colblock++) 
		{
			// Storing blocks as [colblock][rowblock] (i.e. [x][y])
		   	int index_block = colblock + rowblock * colblocks_in;
		   	real *output_block_ = &coldelta[index_block * ip_output_size];
		
			for(int i=0 ; i < ip_output_size ; i++) output_block_[i] = 0.0;
			
			for(int k=-delta_neighbour ; k <= delta_neighbour ; k++) 
			{
		   		int index_convolve_block = colblock + k + rowblock * colblocks_in;
				real *convolve_output_block_ = &dct[index_convolve_block*ip_output_size];

				for(int i=0 ; i < ip_output_size ; i++) 
					output_block_[i] += real(k) * convolve_output_block_[i];
			}
		}
	}

	// vertical delta
	for(int colblock=0; colblock<colblocks_in; colblock++)
	{
	   	for(int rowblock=delta_neighbour; rowblock<(rowblocks_in-delta_neighbour); rowblock++) 
		{
			// Storing blocks as [colblock][rowblock] (i.e. [x][y])
		   	int index_block = colblock + rowblock * colblocks_in;
		   	real *output_block_ = &rowdelta[index_block * ip_output_size];
		
			for(int i=0 ; i < ip_output_size ; i++) output_block_[i] = 0.0;
			
			for(int k=-delta_neighbour ; k <= delta_neighbour ; k++) 
			{
		   		int index_convolve_block = colblock + (rowblock + k) * colblocks_in;
				real *convolve_output_block_ = &dct[index_convolve_block*ip_output_size];

				for(int i=0 ; i < ip_output_size ; i++) 
					output_block_[i] += real(k) * convolve_output_block_[i];
			}
		}
	}
	   	
	// combine horizontal and vertical deltas
	int dump_block = 0;
	
	for(int rowblock=delta_neighbour; rowblock<(rowblocks_in-delta_neighbour); rowblock++) 
	   	for(int colblock=delta_neighbour; colblock<(colblocks_in-delta_neighbour); colblock++) 
		{
			// Storing blocks as [colblock][rowblock] (i.e. [x][y])
		   	int index_block = (colblock + rowblock * colblocks_in) * ip_output_size;
			real *output_block_dct = &dct[index_block];
		   	real *output_block_rowdelta = &rowdelta[index_block];
		   	real *output_block_coldelta = &coldelta[index_block];

			// Keep all DCT coeff except c0, c1 and c2
			for(int i=0 ; i < ip_output_size - delta_keep ; i++)
			   	tmp[i] = output_block_dct[i+delta_keep];

			// Keep some vertical deltas
			for(int i=1 ; i <= delta_keep ; i++)
				tmp[ip_output_size-delta_keep +i -1] = output_block_rowdelta[i -1];
			
			// Keep some horizontal deltas
			for(int i=1 ; i <= delta_keep ; i++)
				tmp[ip_output_size-delta_keep +i +delta_keep -1] = output_block_coldelta[i -1];
			
			for(int i=0 ; i < output_size ; i++)
				seq_out->frames[dump_block][i] = tmp[i];
			
			dump_block++;
		}

	if(dump_block != n_block) 
		warning("ipDCTmod2::process() dump_block != n_block");
}

ipDCTmod2::~ipDCTmod2()
{
}

}

