#include "ipGabor2D.h"
#include <math.h>
#include "ImageGray.h"

namespace Torch {

ipGabor2D::ipGabor2D(int width_, int height_, int s_, int n_, real Ul, real Uh, int scale, int orientation, int verbose_) : ipCore("gray")
{
   	width = width_;
	height = height_;
   	
	s = s_;
	n = n_;

	fft = NULL;
	ifft = NULL;

	if((width < 3) || (height < 3)) error("ipGabor2D(): Image size should not be smaller than 3.");

	seq_out = new(allocator) Sequence(2, width * height);

	Gr = new(allocator) Mat(height, width);
	Gi = new(allocator) Mat(height, width);

	double base, a, u0, z, Uvar, Vvar, Xvar, Yvar, X, Y, G, t1, t2;
	int x, y;

	base = Uh/Ul;
	a = pow(base, 1.0/(double)(scale-1));

	u0 = Uh/pow(a, (double) scale-s);

	Uvar = (a-1.0)*u0/((a+1.0)*sqrt(2.0*log(2.0)));

	z = -2.0*log(2.0)*(Uvar*Uvar)/u0;
	Vvar = tan(M_PI/(2*orientation))*(u0+z)/sqrt(2.0*log(2.0)-z*z/(Uvar*Uvar));

        Xvar = 1.0/(2.0*M_PI*Uvar);
        Yvar = 1.0/(2.0*M_PI*Vvar);

	t1 = cos(M_PI/orientation*(n-1.0));
	t2 = sin(M_PI/orientation*(n-1.0));

	side = (int) (height-1)/2;

	if(verbose_ >= 1)
	{
		print("ipGabor2D::ipGabor2D()\n");
		print("  width = %d\n", width);
		print("  height = %d\n", height);
		print("  s = %d / %d\n", s, scale);
		print("  n = %d / %d\n", n, orientation);
		print("  Ul = %g\n", Ul);
		print("  Uh = %g\n", Uh);
		print("  side = %d\n", side);
	}
	   
	for(x=0 ; x < 2*side+1 ; x++) 
	{
		for(y=0 ; y < 2*side+1 ; y++) 
		{
			X = (double) (x-side)*t1+ (double) (y-side)*t2;
			Y = (double) -(x-side)*t2+ (double) (y-side)*t1;
			G = 1.0/(2.0*M_PI*Xvar*Yvar)*pow(a, (double) scale-s)*exp(-0.5*((X*X)/(Xvar*Xvar)+(Y*Y)/(Yvar*Yvar)));
			Gr->ptr[x][y] = G*cos(2.0*M_PI*u0*X);
			Gi->ptr[x][y] = G*sin(2.0*M_PI*u0*X);
		}
	}

	real *dst_r = seq_out->frames[0];
	real *dst_i = seq_out->frames[1];

	for(int x=0 ; x < height ; x++) 
	{
		for(int y=0 ; y < width ; y++) 
		{
			*dst_r = Gr->ptr[x][y];
			*dst_i = Gi->ptr[x][y];

			dst_r++;
			dst_i++;
		}
	}
			
	if(verbose_ >= 2)
	{
		print("Gabor filter (real part)\n");
		printM(Gr);

		print("Gabor filter (imag part)\n");
		printM(Gi);
	}
}

void seq2mat_(Mat *A, Mat *B, real *a, real *b)
{
  	real *a_ = a;
   	real *b_ = b;

	for(int x=0 ; x < A->m ; x++) 
		for(int y=0 ; y < A->n ; y++) 
		{
			A->ptr[x][y] = *a_++;
			B->ptr[x][y] = *b_++;
		}
}
	
void mat2seq_(Mat *A, Mat *B, real *a, real *b)
{
  	real *a_ = a;
   	real *b_ = b;

	for(int x=0 ; x < A->m ; x++) 
		for(int y=0 ; y < A->n ; y++) 
		{
			*a_++ = A->ptr[x][y];
			*b_++ = B->ptr[x][y];
		}
}

void ipGabor2D::setFFT(ipFFT2D *fft_, ipFFT2D *ifft_, int real_width_, int real_height_)
{
   	fft = fft_;
	ifft = ifft_;
	real_width = real_width_;
	real_height = real_height_;
	
	G_real = new(allocator) Mat(fft->height, fft->width);
	G_imag = new(allocator) Mat(fft->height, fft->width);
	I_real = new(allocator) Mat(fft->height, fft->width);
	I_imag = new(allocator) Mat(fft->height, fft->width);
	F_real = new(allocator) Mat(fft->height, fft->width);
	F_imag = new(allocator) Mat(fft->height, fft->width);
	tmp_r = new(allocator) Mat(fft->height, fft->width);
	tmp_i = new(allocator) Mat(fft->height, fft->width);
	seq_ = new(allocator) Sequence(2, fft->height * fft->width);

	allocator->free(seq_out);
	seq_out = new(allocator) Sequence(2, real_height * real_width);
}

void ipGabor2D::process(Sequence *seq_in)
{
	if(fft != NULL)
	{
		if(seq_in->frame_size != fft->width * fft->height) warning("ipGabor2D::process() incorrect frame size (%d) for FFT.", seq_in->frame_size);
		else
		{
		   	if(verbose)
			{
		   		print("ipGabor2D::process() using FFT\n");
			   	print("   filter width = %d\n", width);
			   	print("   filter height = %d\n", height);
			   	print("   FFT width = %d\n", fft->width);
			   	print("   FFT height = %d\n", fft->height);
			   	print("   IFFT width = %d\n", ifft->width);
			   	print("   IFFT height = %d\n", ifft->height);
			   	print("   real width = %d\n", real_width);
			   	print("   real height = %d\n", real_height);
			   	print("   side = %d\n", side);
			}

			//
			// copy the input sequence into a matrix
		   	if(verbose) print("FFT input image\n");

			I_real->zero();
			I_imag->zero();
			seq2mat_(I_real, I_imag, seq_in->frames[0], seq_in->frames[1]);
			

			//
			// compute the FFT of the Gabor filter
		   	if(verbose) print("FFT gabor filter\n");

			// copy the Gabor filter into a matrix of size power of 2
			tmp_r->zero();
			tmp_i->zero();

			mxCopy(tmp_r, Gr, 0, 0, 0, 0, 2*side, 2*side);
			mxCopy(tmp_i, Gi, 0, 0, 0, 0, 2*side, 2*side);

			// copy the new Gabor filter matrix into a sequence
			mat2seq_(tmp_r, tmp_i, seq_->frames[0], seq_->frames[1]);

			// compute the FFT
			fft->process(seq_);

			// copy the FFT into a matrix
			G_real->zero();
			G_imag->zero();
			seq2mat_(G_real, G_imag, fft->seq_out->frames[0], fft->seq_out->frames[1]);


			//
			// filter the image I with the Gabor filter G
			// Nb: in the image domain it should be a convolution
			// but here in the FT domain this is just a
			// multiplication of complex numbers GxI
			if(verbose) print("Filtering\n");

			// compute the real part of the filtered image F: F.r = G.r x I.r - G.i x I.i
			tmp_r->zero();
			tmp_i->zero();
			F_real->zero();
			mxMulMat(G_real, I_real, tmp_r);
			mxMulMat(G_imag, I_imag, tmp_i);
			mxMatSubMat(tmp_r, tmp_i, F_real);
			
			// compute the imag part of the filtered image F: F.i = G.r x I.i + G.i x I.r
			tmp_r->zero();
			tmp_i->zero();
			F_imag->zero();
			mxMulMat(G_real, I_imag, tmp_r);
			mxMulMat(G_imag, I_real, tmp_i);
			mxMatAddMat(tmp_r, tmp_i, F_imag);
		
			// copy the filtered image into a sequence
			mat2seq_(F_real, F_imag, seq_->frames[0], seq_->frames[1]);

			// compute the IFFT of the filtered image
			ifft->process(seq_);

			// copy the filtered image
			cropGray(ifft->seq_out->frames[0], fft->width, fft->height, 2*side, 2*side, real_width, real_height, seq_out->frames[0]);
			cropGray(ifft->seq_out->frames[1], fft->width, fft->height, 2*side, 2*side, real_width, real_height, seq_out->frames[1]);
		}	
	}
	else
	{
		if(seq_in->frame_size != width * height) warning("ipGabor2D::process() incorrect frame size. You should perform a convolution.");
		else
		{
			real *src_r = seq_in->frames[0];
			real *dst_r = seq_out->frames[0];
			real *dst_i = seq_out->frames[1];

			for(int x=0 ; x < height ; x++) 
			{
				for(int y=0 ; y < width ; y++) 
				{
				   	real r_ = *src_r;
					
					*dst_r = r_ * Gr->ptr[x][y];
					*dst_i = r_ * Gi->ptr[x][y];

					src_r++;
					dst_r++;
					dst_i++;
				}
			}
		}
	}
}

ipGabor2D::~ipGabor2D()
{
}

}

