#ifndef VISION_INC
#define VISION_INC

#include "Object.h"
#include "Sequence.h"

namespace Torch {

/** RGB pixel

    \begin{verbatim}
    r is the red component
    g is the green component
    b is the blue component
    \end{verbatim}

    @author Sebastien Marcel (marcel@idiap.ch)
    @version 2.0
    \Date
    @since 1.0
*/
struct sPixRGB 
{
	unsigned char r, g, b;
};

/** YUV pixel

    \begin{verbatim}
    y is the Y component
    u is the U component
    v is the V component
    \end{verbatim}

    @author Sebastien Marcel (marcel@idiap.ch)
    @version 2.0
    \Date
    @since 1.0
*/
struct sPixYUV
{
	unsigned char y, u, v;
};

/** RGB pixel (in real)

    \begin{verbatim}
    r is the red component
    g is the green component
    b is the blue component
    \end{verbatim}

    @author Sebastien Marcel (marcel@idiap.ch)
    @version 2.0
    \Date
    @since 1.0
*/
struct rgbReal
{
	real r, g, b;
};

/** YUV pixel (in real)

    \begin{verbatim}
    y is the Y component
    u is the U component
    v is the V component
    \end{verbatim}

    @author Sebastien Marcel (marcel@idiap.ch)
    @version 2.0
    \Date
    @since 1.0
*/
struct yuvReal
{
	real y, u, v;
};

/** Point2D (in double)

    \begin{verbatim}
    x is the x coordinate
    y is the y coordinate
    \end{verbatim}

    @author Sebastien Marcel (marcel@idiap.ch)
    @version 2.0
    \Date
    @since 1.0
*/
struct sPoint2D
{
	double x, y;
};

/** Complex (in double)

    \begin{verbatim}
    r is the real part
    i is the imaginary part
    \end{verbatim}

    @author Sebastien Marcel (marcel@idiap.ch)
    @version 2.0
    \Date
    @since 2.0
*/
struct sComplex
{
	real r, i;
};

#define FixI(v) (int) (v+0.5)
#define IS_NEAR(var, value, delta) ((var >= (value - delta)) && (var <= (value + delta)))

/// This function returns the index of a two-dimensional image represented as one-dimensional 
int iIndex(int y, int x, int width);

/// This function returns the log2
double log2(double a);

/** Crop and Paste functions 

    @author Sebastien Marcel (marcel@idiap.ch)
    @version 2.0
    \Date
    @since 1.0
*/
//@{

/** Crop a subimage from gray level image

    \begin{verbatim}

         src                            dest
           +--------------------+         +----+ 
	   |                    |         |XXXX| h
	   |                    |         |XXXX|
	   |      x,y           |         +----+
	   |       +----+       |           w
           |       |XXXX| h     |
           |       |XXXX|       | height
	   |       +----+       |
	   |         w          |
	   |                    |
	   |                    |
	   |                    |
	   |                    |
	   |                    |
	   |                    |
	   |                    |
           +--------------------+
                   width
	   
    \end{verbatim}

    @param src_ is the source image
    @param width_ is the #width# of the source image
    @param height_ is the #height# of the source image
    @param x_ is the upper left #x# coordinate of the subwindow to crop
    @param y_ is the upper left #y# coordinate of the subwindow to crop
    @param w_ is the #width# of the subwindow to crop
    @param h_ is the #height# of the subwindow to crop
    @param dest_ is the cropped subwindow
*/
bool cropGray(real *src_, int width_, int height_, int x_, int y_, int w_, int h_, real *dest_);
	
/** Crop a subimage from rgb image

    @param src_ is the source image
    @param width_ is the #width# of the source image
    @param height_ is the #height# of the source image
    @param x_ is the upper left #x# coordinate of the subwindow to crop
    @param y_ is the upper left #y# coordinate of the subwindow to crop
    @param w_ is the #width# of the subwindow to crop
    @param h_ is the #height# of the subwindow to crop
    @param dest_ is the cropped subwindow

    @see cropGray
*/
bool cropRgb(real *src_, int width_, int height_, int x_, int y_, int w_, int h_, real *dest_);

/** Paste a gray level subimage at a given location into another image

    \begin{verbatim}

         src            dest
           +----+         +--------------------+ 
	   |XXXX| h       |                    | 
	   |XXXX|         |                    | 
	   +----+         |      x,y           | 
	     w            |       +----+       | 
                          |       |XXXX| h     |
                          |       |XXXX|       | height
	                  |       +----+       |
	                  |         w          |
	                  |                    |
	                  |                    |
                          +--------------------+
                                 width
	   
    \end{verbatim}

    @param src_ is the source image (subwindow)
    @param x_ is the upper left #x# coordinate where to paste
    @param y_ is the upper left #y# coordinate where to paste
    @param w_ is the #width# of the source image
    @param h_ is the #height# of the source image
    @param width_ is the #width# of the destination image
    @param height_ is the #height# of the destination image
    @param dest_ is the destination image which will be pasted
*/
bool pasteGray(real *src_, int x_, int y_, int w_, int h_, real *dest_, int width_, int height_);

//@}

/// This function rescale an image between 0 and #scale_#
void scaleGray(int w, int h, real *image, real scale_ = 255.0);

}

#endif 
