#include "EigenMachine.h"

#include "DiskXFile.h"

namespace Torch {

EigenMachine::EigenMachine(int n_inputs_)
{
   	n_inputs = n_inputs_;
	n_outputs = n_inputs;
	
   	//
	outputs = new(allocator) Sequence(1, n_outputs);

	//
    	eigenvalues = NULL;
  	eigenvalues = (real *) allocator->alloc(sizeof(real) * n_inputs);

	//
    	eigenvectors = NULL;
	real *base_ = (real *)allocator->alloc(sizeof(real) * n_inputs * n_inputs);
        eigenvectors = (real **)allocator->alloc(sizeof(real *) * n_inputs);
        for(int i = 0; i < n_inputs ; i++) eigenvectors[i] = base_ + i * n_inputs;

	//
	addROption("variance", &variance, 1.0, "variance to keep");
	addIOption("verbose_level", &verbose_level, 0, "verbose level");
}

void EigenMachine::init()
{
   	if(variance <= 0.0)
	{
	   	n_outputs = n_inputs;
		if(verbose_level >= 1) print("Keeping all eigenvectors: %d\n", n_outputs);
	}
	else
	{
   		int i;
   		real sum = eigenvalues[0];
   		real variance_ = 0.0;

		real min_eigenv = eigenvalues[0];
		real max_eigenv = eigenvalues[0];

		if(verbose_level >= 2)
			print(" eigen value [0] = %g\n", eigenvalues[0]);
		
   		for(i = 1 ; i < n_inputs ; i++)
		{
			if(verbose_level >= 2)
				print(" eigen value [%d] = %g\n", i, eigenvalues[i]);

			if(eigenvalues[i] < min_eigenv) min_eigenv = eigenvalues[i];
			if(eigenvalues[i] > max_eigenv) max_eigenv = eigenvalues[i];
			
		   	sum += eigenvalues[i];
		}
		
		if(verbose_level >= 1)
		{
			print("Total variance = %g\n", sum);
			print("Eigenvalue max = %g\n", max_eigenv);
			print("Eigenvalue min = %g\n", min_eigenv);
		}
		
   		i = 0;
   		while (variance_ < (sum*variance))
   		{
      			variance_ += eigenvalues[i];
      			i++;
   		}

		n_outputs = i;

		if(verbose_level >= 1) print("%g per cent of variance => %d Eigenvectors selected\n", variance, n_outputs);
	}

	if(verbose_level >= 2)
	{
		print("[ ");
		for(int i = 0 ; i < n_inputs ; i++)
			print("%g ", eigenvalues[i]);
		print("]\n");
	}

}

void EigenMachine::forward(Sequence *inputs)
{
   	if(verbose_level >= 3) print("EigenMachine::forward() %d->%d ...\n", n_inputs, n_outputs);
	
	for(int i = 0 ; i < n_outputs ; i++)
	{
	   	outputs->frames[0][i] = 0.0;

       		for(int j = 0 ; j < n_inputs ; j++)
           		outputs->frames[0][i] += eigenvectors[j][i] * inputs->frames[0][j];
	}
}

void EigenMachine::loadXFile(XFile *file)
{
   	if(verbose_level >= 2) print("EigenMachine::loadXFile() ...\n");

	int n_inputs_;
	
   	file->taggedRead(&n_inputs_, sizeof(int), 1, "N_INPUTS");
	if(n_inputs_ != n_inputs) error("EigenMachine::loadXFile() n_input is incorrect.");
	
	file->taggedRead(eigenvalues, sizeof(real), n_inputs, "EIGENVALUES");
	for(int row = 0 ; row < n_inputs ; row++)
		file->taggedRead(eigenvectors[row], sizeof(real), n_inputs, "EIGENVECTORS");
}

void EigenMachine::saveXFile(XFile *file)
{
   	if(verbose_level >= 2) print("EigenMachine::saveXFile() ...\n");
	
   	file->taggedWrite(&n_inputs, sizeof(int), 1, "N_INPUTS");
	file->taggedWrite(eigenvalues, sizeof(real), n_inputs, "EIGENVALUES");
	for(int row = 0 ; row < n_inputs ; row++)
		file->taggedWrite(eigenvectors[row], sizeof(real), n_inputs, "EIGENVECTORS");
}

void EigenMachine::load(const char *filename, PreProcessing *preprocess_norm)
{
   	if(verbose_level >= 2) print("EigenMachine::load(%s) ...\n", filename);
	
   	DiskXFile *file = NULL;

	file = new DiskXFile(filename, "r");
	
	//
	loadXFile(file);

	//
	if(preprocess_norm != NULL)
                preprocess_norm->loadXFile(file);

	delete file;
}

void EigenMachine::save(const char *filename, PreProcessing *preprocess_norm)
{
   	if(verbose_level >= 2) print("EigenMachine::save(%s) ...\n", filename);
	
   	DiskXFile *file = NULL;

	file = new DiskXFile(filename, "w");
	
	//
	saveXFile(file);
	
	//
	if(preprocess_norm != NULL)
                preprocess_norm->saveXFile(file);

	delete file;
}

EigenMachine::~EigenMachine()
{
}

}
