#include "PatternDetect.h"

namespace Torch {

PatternDetect::PatternDetect(	int width_image_, int height_image_, 
				int patternWmin_, int patternHmin_, 
				int Wmin_, int Wmax_, real stepx_, real stepy_,
				real scale_factor_,
				real minMean_, real maxMean_, real minStdv_, real maxStdv_,
				real *pruning_ii_,
				real pruning_min_threshold_,
				real pruning_max_threshold_,
				int max_pattern_,
				bool verbose_)
{
	verbose = verbose_;

	width_image = width_image_;
        height_image = height_image_;
        ratioImage = (real) height_image / width_image;

	patternWmin = patternWmin_;
        patternHmin = patternHmin_;
        ratioPattern = (real) patternHmin / patternWmin;
	
	if(scale_factor_ <= 0) 
		error("scale factor (%g) must be positive!", scale_factor_);
	scale_factor = 1.0 + scale_factor_;
	stepx_factor = stepx_;
        stepy_factor = stepy_;
	

	/****** compute Wmin/Wmax  ******/
        Wmin = (Wmin_ > patternWmin ? Wmin_ : patternWmin);
        Wmax = (Wmax_ < (real)Wmin/ratioPattern ? width_image : Wmax_);
	
	if(Wmax < Wmin) error("Wmax (%d) must be >= to Wmin (%d)!", Wmax, Wmin);

        real w_factor = (real)width_image/patternWmin;
        real h_factor = (real)height_image/patternHmin;

        if(h_factor > w_factor)
                Wmax = (Wmax < width_image ? Wmax : width_image);
        else
        {
                int Hmax = height_image;
                int Wmax__ = FixI((real)Hmax * ratioPattern);
                Wmax = (Wmax < Wmax__ ? Wmax : Wmax__);
        }

	/****** Determine the number of scale levels ******/
        int width_pattern_ = Wmin;
        n_levels = 0;
        while(width_pattern_ <= Wmax)
        {
                width_pattern_ = FixI((real) width_pattern_ * scale_factor);
                n_levels++;
        }

	levels = (WindowScan **)allocator->alloc(sizeof(WindowScan *)*n_levels);
	for(int i = 0 ; i < n_levels ; i++) levels[i] = NULL;
	
	ip_ii = new(allocator) ipIntegralImage(width_image, height_image, "gray", true);
	
	ip_sw = NULL,

	patterncollector = new(allocator) PatternCollector(patternWmin, patternHmin, max_pattern_);
	
	minMean = minMean_;
	maxMean = maxMean_;
	minStdv = minStdv_;
	maxStdv = maxStdv_;
	
	pruning_ii = pruning_ii_;
	pruning_min_threshold = pruning_min_threshold_;
	pruning_max_threshold = pruning_max_threshold_;
	
	init_done = false;

	n_patterns = 0;
	patterns = NULL;

	n_subwindows = 0;
	n_subwindows_processed = 0;
	n_pruning_reject = 0;
	n_mean_reject = 0;
	n_stdv_reject = 0;

	if(verbose)
	{
		print(" model size: %dx%d\n", patternWmin, patternHmin);
                print(" scanning:   Wmin =  %d\n", Wmin);
                print("             Wmax =  %d\n", Wmax);
		print("Number of scales computed = %d\n", n_levels);		
	}
}

void PatternDetect::init(void)
{
	for(int i = 0; i < n_levels; i++)
	{
	   	levels[i]->setMeanStdv(minMean, maxMean, minStdv, maxStdv, ip_ii->seq_out->frames[0], ip_ii->seq_out->frames[1]);
	   	levels[i]->setPruningII(pruning_min_threshold, pruning_max_threshold, pruning_ii);
		
		levels[i]->init();
	}
	
	init_done = true;
}

void PatternDetect::process(Sequence *seq_in)
{
	n_subwindows = 0;
	n_subwindows_processed = 0;
	n_pruning_reject = 0;
        n_mean_reject = 0;
        n_stdv_reject = 0;

	if(init_done)
	{
		patterncollector->reset();

	   	ip_ii->process(seq_in);

		for(int i = 0 ; i < n_levels ; i++)
		{
			levels[i]->process(seq_in);

			n_subwindows += levels[i]->n_subwindows;
			n_subwindows_processed += levels[i]->n_subwindows_processed;
			n_pruning_reject += levels[i]->n_pruning_reject;
			n_mean_reject += levels[i]->n_mean_reject;
			n_stdv_reject += levels[i]->n_stdv_reject;
		}

		mergepatterns();
	}
	else
		warning("PatternDetect::process -> do init() first.");
}

void PatternDetect::stat()
{
	print("\n------------  PatternDetect::stat()  -----------\n");
	print("                                        \n"); 
	print("    number of subwindows : %d\n", n_subwindows);
	print("    subwindows processed : %d\n", n_subwindows_processed);
	print("                                        \n");
	print("    pruning rejections   : %d\n", n_pruning_reject);
	print("    mean rejections      : %d\n", n_mean_reject);
	print("    stdv rejections      : %d\n", n_stdv_reject);
	print("                                        \n");
	
	ip_sw->stat();

	print("    detections after fusion  : %d\t(%d)\n", n_patterns, patterncollector->n_patterns);
	print("                                        \n");
	print("---------------------------------------------------\n\n");
}

PatternDetect::~PatternDetect()
{
}

}
