#include "PyramidWindowScan.h"

namespace Torch {

real iisum(real *integralimage_, int width_, int height_, int x_window_, int y_window_, int width_window_, int height_window_);
	
PyramidWindowScan::PyramidWindowScan(	
		int width_image_, int height_image_, 
		int real_width_image_, int real_height_image_, 
		int width_pattern_, int height_pattern_, 
		int real_width_pattern_, int real_height_pattern_,
		real stepx_factor_, real stepy_factor_,
		ipSubWindow *ip_sw_, 
		PatternCollector *patterncollector_,
		real scale_,
		bool verbose)
		: WindowScan(width_image_, height_image_, real_width_image_, real_height_image_, width_pattern_, height_pattern_, real_width_pattern_, real_height_pattern_, stepx_factor_, stepy_factor_, ip_sw_, patterncollector_, verbose)
{
	if(verbose)
	{
		print("PyramidWindowScan::PyramidWindowScan() scan %dx%d image (real=%dx%d) with %dx%d window (real=%dx%d)\n", width_image, height_image, real_width_image, real_height_image, width_pattern, height_pattern, real_width_pattern, real_height_pattern);
		print(" + stepx = %d\n", stepx);
		print(" + stepy = %d\n", stepy);
	}
	
	scale = scale_;

	if(!IS_NEAR(scale, (real) real_width_pattern / width_pattern, 0.001))
		error("WindowScan: error with the scale (%g != %g)\n", scale, (real) real_width_pattern / width_pattern);
	
	subsampling = NULL;
	subsample_image = NULL;
	
	if(scale != 1.0)
	{
		subsampling = new(allocator) ipScale_yx(real_width_image, real_height_image, "gray", width_image, height_image);
		subsample_image = subsampling->seq_out;
	}
}

void PyramidWindowScan::init(void)
{
	ip_sw->init();
}

void PyramidWindowScan::process(Sequence *seq_in)
{
	// set the subsampled image size
	ip_sw->width_image = width_image;
	ip_sw->height_image = height_image;

	// set the subwindow size
	// rem.: could be done in init, because constant
	ip_sw->w_window = width_pattern;
	ip_sw->h_window = height_pattern;

	if(subsampling) subsampling->process(seq_in);
	else subsample_image = seq_in;
	
	//
	int scan_width = width_image - width_pattern;
	int scan_height = height_image - height_pattern;

	//
	int y;
	int x;

	//
	n_subwindows = 0;
	n_subwindows_processed = 0;
	n_pruning_reject = 0;
	n_mean_reject = 0;
	n_stdv_reject = 0;
	
	y = 2;
	
	while(y < scan_height-1)
	{
   	    x = 2;

	    while(x < scan_width-1)
            {
		n_subwindows++;
		    
		int x_ = FixI(scale * x);
		int y_ = FixI(scale * y);

		// PRUNING FILTER
		if(use_pruning)
		{
			real pruning_sum = iisum(pruning_ii, real_width_image, real_height_image, x_, y_, real_width_pattern, real_height_pattern);

			if((pruning_sum < pruning_min_threshold) || (pruning_sum > pruning_max_threshold))
			{
				x += stepx;
				n_pruning_reject++;
				continue;
			}
		}

		// MEAN-STDV FILTER
		if(use_mean_stdv)
		{
		    real mean = iisum(integralimage, real_width_image, real_height_image, x_, y_, real_width_pattern, real_height_pattern);

		    if((mean < minMean) || (mean > maxMean))
		    {
			    x += stepx;
			    n_mean_reject++;
			    continue;
		    }

		    real stdv = iisum(integralimage2, real_width_image, real_height_image, x_, y_, real_width_pattern, real_height_pattern);

   	            stdv -= mean*mean*one_by_size_pattern;

		    if((stdv < minStdv) || (stdv > maxStdv))	    
		    {
			    x += stepx;
			    n_stdv_reject++;
			    continue;
		    }
		}

		// AND FINALLY THE MACHINE  
		// set the window location
		ip_sw->x_window = x;
		ip_sw->y_window = y;

		// process the subwindow
		ip_sw->process(subsample_image);

		//if(ip_sw->seq_out->frames[0][0] > ip_sw->threshold)
		if(ip_sw->isPattern())
		{
		    // add a new pattern
		    if(patterncollector) 
			patterncollector->add(x_, y_, real_width_pattern, real_height_pattern, ip_sw->seq_out->frames[0][0], ip_sw->view);
		}
		
		n_subwindows_processed++;
		
		x += stepx;
	    } 
	    y += stepy;
	}
}

PyramidWindowScan::~PyramidWindowScan()
{
}

}

