#include "rgbRawVideoFile.h"
#include "Convert.h"

namespace Torch {

#undef FixI
#define FixI(v) (int) (v+0.5)
   
rgbRawVideoFile::rgbRawVideoFile()
{
#ifndef HAVE_AVILIB
	error("HAVE_AVILIB not defined in rgbRawVideoFile::rgbRawVideoFile()");
#endif
   
	readMode = false;
	writeMode = false;

	addIOption("width", &width, 576, "output width");
	addIOption("height", &height, 576, "output height");
	addROption("framerate", &frate, 25, "output framerate");

#ifdef HAVE_AVILIB
	avifile_input = NULL;
	avifile_output = NULL;
#endif
}
  	
rgbRawVideoFile::rgbRawVideoFile(char *filename_, const char *open_flags_) : VideoFile(filename_, open_flags_)
{
#ifndef HAVE_AVILIB
	error("HAVE_AVILIB not defined in rgbRawVideoFile::rgbRawVideoFile()");
#endif
   
	readMode = false;
	writeMode = false;

	addIOption("width", &width, 576, "output width");
	addIOption("height", &height, 576, "output height");
	addROption("framerate", &frate, 25, "output framerate");

#ifdef HAVE_AVILIB
	avifile_input = NULL;
	avifile_output = NULL;
#endif
}

bool rgbRawVideoFile::open(char *filename_, const char *open_flags_)
{
	readMode = false;
	writeMode = false;

	cframe = 0;
	
#ifdef HAVE_AVILIB
	if(strcmp(open_flags_, "r") == 0) 
	{
				
		// open the file in read mode
		avifile_input = AVI_open_input_file(filename_, 1);

		if (avifile_input == NULL)
		{
			print("Error opening file %s\n",filename_);

			return false;
		}

		readMode = true;

		// get the width
		width = AVI_video_width(avifile_input);

		// get the height
		height = AVI_video_height(avifile_input);

		// get the frame rate
		frate = AVI_frame_rate(avifile_input);

		// get the number of images
		nframes = AVI_video_frames(avifile_input);

		// get the codec name
		codec_name =  AVI_video_compressor(avifile_input);

		// allocate a new pixmap of appropriate size
		size = width * height * 3;
		if(pixmap == NULL) pixmap = (unsigned char *) allocator->alloc(size * sizeof(unsigned char));
		pixmap = (unsigned char *) allocator->realloc(pixmap, size * sizeof(unsigned char));
	}
	else if(strcmp(open_flags_, "w") == 0) 
	{
		// open the file in write mode
		avifile_output = AVI_open_output_file(filename_);

		if (avifile_output == NULL)
		{
			print("Error creating file %s\n",filename_);
			return false;
		}

		writeMode = true;
		
		codec_name = NULL;
	}

	return true;
#else
	warning("rgbRawVideoFile::open() avilib not supported.");

	return false;
#endif
}

bool rgbRawVideoFile::read()
{
   	if(readMode == false)
	{	
	   	warning("Impossible to read (not in read mode)");

		return false;
	}
	
	if(cframe > nframes) 
	{
		warning("End of file.\n");
		
		return false;
	}
	
#ifdef HAVE_AVILIB	
	int key;

	// read the current frame (should be flipped BGR24)
	AVI_read_frame(avifile_input, (char *) pixmap, &key);

	// flip and convert a BGR24 into a RGB24
	rgb24_to_bgr24_flip(pixmap, NULL, width, height);
	
	cframe++;
	
	return true;
#else
	warning("rgbRawVideoFile::read() avilib not supported.");

	return false;
#endif
}

bool rgbRawVideoFile::write(unsigned char *pixmap_)
{
   	if(writeMode == false)
	{	
	   	warning("Impossible to write (not in write mode)");

		return false;
	}
	
#ifdef HAVE_AVILIB	
	if(cframe == 0) // This is the first frame
	{
		size = width * height * 3;

		// allocate a new pixmap of appropriate size
		if(pixmap == NULL) pixmap = (unsigned char *) allocator->alloc(size * sizeof(unsigned char));
		pixmap = (unsigned char *) allocator->realloc(pixmap, size * sizeof(unsigned char));

		// write video header
		AVI_set_video(avifile_output, width, height, frate, "RGB" );
	}	

	// flip and convert a RGB24 (#pixmap_#) into a BGR24 (#pixmap#)
	rgb24_to_bgr24_flip(pixmap_, pixmap, width, height);
	
	// 
	int ret = AVI_write_frame(avifile_output, (char *)pixmap, size, 1);

	cframe++;

	if(ret) 
	{
		warning("Impossible to write frame (%d)\n", cframe);
		
		return false;
	}
	
	return true;
#else
	warning("rgbRawVideoFile::write() avilib not supported.");

	return false;
#endif
}  

bool rgbRawVideoFile::rewindmovie()
{
   	if(readMode == false)
	{	
	   	warning("Impossible to rewind (not in read mode)");

		return false;
	}
	
#ifdef HAVE_AVILIB	
	// starts the video at the beginning
	AVI_seek_start(avifile_input);
	
	return true;
#else
	warning("rgbRawVideoFile::rewindmovie() avilib not supported.");

	return false;
#endif
}

int rgbRawVideoFile::seekmovie(int h, int m, int s, int f)
{
   	if(readMode == false)
	{	
	   	warning("Impossible to seek (not in read mode)");

		return -1;
	}
	
#ifdef HAVE_AVILIB	
	long timeToSeek;
	
	// computes the position to seek
	timeToSeek = FixI((h*3600+m*60+s)*frate) + f;

	// seek to the position #timeToSeek# into the file
	AVI_set_video_position(avifile_input, timeToSeek);

	return timeToSeek;
#else
	warning("rgbRawVideoFile::seekmovie() avilib not supported.");

	return -1;
#endif
}

bool rgbRawVideoFile::close()
{
#ifdef HAVE_AVILIB	
   	//
	if(readMode) AVI_close(avifile_input);
	if(writeMode) AVI_close(avifile_output);
	
	//
	if(pixmap != NULL) allocator->free(pixmap);
	pixmap = NULL;

	return true;
#else
	warning("rgbRawVideoFile::close() avilib not supported.");

	return false;
#endif
}

rgbRawVideoFile::~rgbRawVideoFile()
{
	if(pixmap != NULL) allocator->free(pixmap);
	pixmap = NULL;
}

}

